local L = AceLibrary("AceLocale-2.2"):new("GhostPulse")

local pairs = _G.pairs


ParryWatcher = {

	New = function(this, env, listenerFunction)
		-- check arguments
		if not env or type(env) ~= "table" then
			error("Must supply an environment tracker to ParryWatcher:New()")
		end
		
		if not listenerFunction or type(listenerFunction) ~= "function" then
			error("Must supply a listener function to ParryWatcher:New()")
		end
			
		-- globals
		local frame = CreateFrame("Frame")
		
		
		frame.watcherName = "Parry"
		
		frame.enabled = false		
		frame.enabledAlerts = { }
				
		frame.environment = env
		frame.OnDetect = listenerFunction

		-- functions
		frame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
		
		frame.GetPlayerInfo = function(this)
			if this.playerClass == nil or this.playerClass == "" then
				this.playerClass = _, UnitClass("player")
				this.playerGuid = UnitGUID("player")
				this.playerName = UnitName("player")
			end
		end
		
		frame.GetAlertTypes = function()
			return {
				["parryplayer"] = L["Riposte"],
				["parrytarget"] = L["Overpower (Target)"],
				["parryfocus"] = L["Overpower (Focus)"],
			}
		end
		
		-- common configuration
		frame.SetOptions = Environment.SetOptions		
		frame.CheckEnv = Environment.CheckEnv
		frame.CheckCombat = Environment.CheckCombat
		frame.CheckGrouping = Environment.CheckGrouping
		frame.CanShowThing = Environment.CanShowThing		
		frame.SetAlertEnabled = Environment.SetAlertEnabled
		
		-- watcher-specific function
		frame:SetScript("OnEvent", function(this, event, ...)
		if this.enabled then
				if event == "COMBAT_LOG_EVENT_UNFILTERED" then
					if this.enabledAlerts["parryplayer"] or
					   this.enabledAlerts["parrytarget"] or
					   this.enabledAlerts["parryfocus"] then
						this:NotifyParry(...)
					end
				end
			end
		end)
		
		frame.NotifyParry = function(this, timeStamp, event, srcGuid, srcName, srcFlags, destGuid, destName, destFlags, missType)
			this:GetPlayerInfo()

			if event == "SWING_MISSED" and missType == "PARRY" and destGuid == this.playerGuid then
				local _, pClass = UnitClass("player")
				local _, _, _, _, hasRiposte = GetTalentInfo(2, 8)


				-- if we're a rogue, alert a Riposte straight away - the target/focus doesn't matter
				if hasRiposte and pClass == "ROGUE" then
					-- Riposte
					local name, _, riposteTexture = GetSpellInfo(14251)
					local start, duration, enabled = GetSpellCooldown(name, BOOKTYPE_SPELL)
					
					
					if enabled == 1 and duration == 0 then
						this:Alert("player", this.playerName, L["Riposte"], riposteTexture)
					end
				end	
				
			elseif this.playerClass == "WARRIOR" and GetShapeshiftForm(true) == 1 then
				-- overpower if the attacker we parried was the target or the focus
				local suffix = ""


				if srcGuid == UnitGUID("target") then
					suffix = "target"
				elseif srcGuid == UnitGUID("focus") then
					suffix = "focus"
				end
								
				if suffix ~= "" then
					local _, _, overPowerTexture = GetSpellInfo(11585)
				
					
					-- alert
					this:Alert(suffix, srcName, L["Overpower"], overPowerTexture)
				end				
			end
		end

		frame.Alert = function(this, unit, unitName, spellName, icon)
			if this.OnDetect then
				if type(this.OnDetect) == "function" then
					local info = {
						["Unit"] = unit,
						["Name"] = unitName,
						["Spell"] = spellName,
						["Type"] = "parry"..unit,
					}
					
					this:OnDetect(this, "parry"..unit, spellName, icon, info)
				else
					error("ParryWatcher.OnDetect must be a function")
				end
			end
		end		
		
		return frame
	end,
	
}